from typing import NamedTuple, Callable, List

import onnx


class Graph(NamedTuple):
    generate_name: Callable[[], str]
    inputs: List[onnx.ValueInfoProto] = []
    outputs: List[onnx.ValueInfoProto] = []
    transients: List[onnx.ValueInfoProto] = []
    nodes: List[onnx.NodeProto] = []
    initializers: List[onnx.TensorProto] = []


def create_name_generator() -> Callable[[], str]:
    state = {}
    def _generate_unique_name(name: str) -> str:
        """ Generates a new globaly unique name in the graph
        """
        if name in state:
            i = state[name]
            state[name] += 1
        else:
            state[name] = 0
            
        return "{}_{}".format(name, state[name])
        
        
    return _generate_unique_name


def extend(i, val):
    """Extends a list as a copy
    """
    ii = list(i)
    ii.extend(val)
    return ii


def pipe(*args):
    pass


def create_graph():
    return Graph(
        generate_name=create_name_generator()
    )

def compile(graph, target_opset, name="ebm"):
    #outputs = graph.transients

    graph = onnx.helper.make_graph(
        nodes=graph.nodes,
        name=name,
        inputs=graph.inputs,    
        outputs=graph.outputs,
        initializer=graph.initializers,
    )
    model = onnx.helper.make_model(graph, producer_name='ebm2onnx')
    model.opset_import[0].version = target_opset
    return model

def create_input(graph, name, type, shape):
    input = onnx.helper.make_tensor_value_info(name , type, shape)
    return Graph(
        generate_name=graph.generate_name,
        inputs=[input],
        transients=[input],
    )


def add_output(graph, name, type, shape):
    output = onnx.helper.make_tensor_value_info(name , type, shape)
    return graph._replace(
        outputs=extend(graph.outputs, [output]),
    )


def create_initializer(graph, name, type, shape, value):
    initializer = onnx.helper.make_tensor(graph.generate_name(name) , type, shape, value)
    return Graph(
        generate_name=graph.generate_name,
        initializers=[initializer],
        transients=[initializer],
    )


def strip_to_transients(graph):
    """ Returns only the transients of a graph
    """
    return Graph(
        generate_name=graph.generate_name,
        transients=graph.transients,
    )


def clear_transients(graph):
    """ Removes all transients from a graph
    """
    return graph._replace(
        inputs=graph.inputs,
        outputs=graph.outputs,
        initializers=graph.initializers,
        transients=[],
        nodes=graph.nodes,
    )


def merge(*args):
    g = None

    for graph in args:
        if g is None:
            g = graph
        else:
            g = g._replace(
                inputs=extend(g.inputs, graph.inputs),
                outputs=extend(g.outputs, graph.outputs),
                initializers=extend(g.initializers, graph.initializers),
                transients=extend(g.transients, graph.transients),
                nodes=extend(g.nodes, graph.nodes),
            )

    return g


import onnx
import ebm2onnx.graph as graph


def add():
    def _add(g):
        add_result_name = g.generate_name('add_result')
        nodes = [
            onnx.helper.make_node("Add", [g.transients[0].name, g.transients[1].name], [add_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(add_result_name, g.transients[0].type.tensor_type.elem_type, []),
            ],
        )

    return _add


def argmax(axis=0, keepdims=1, select_last_index=0):
    def _argmax(g):        
        argmax_result_name = g.generate_name('argmax_result')
        nodes = [
            onnx.helper.make_node(
                "ArgMax",
                [g.transients[0].name], [argmax_result_name],
                axis=axis, keepdims=keepdims, select_last_index=select_last_index
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(argmax_result_name, onnx.TensorProto.INT64, []),
            ],
        )

    return _argmax


def cast(to):
    def _cast(g):        
        cast_result_name = g.generate_name('cast_result')
        nodes = [
            onnx.helper.make_node("Cast", [g.transients[0].name], [cast_result_name], to=to),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(cast_result_name, to, []),
            ],
        )

    return _cast


def concat(axis):
    def _concat(g):
        concat_result_name = g.generate_name('concat_result')

        sources = [t.name for t in g.transients]
        nodes = [
            onnx.helper.make_node("Concat", sources, [concat_result_name], axis=axis),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(concat_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _concat


def expand():
    def _expand(g):        
        expand_result_name = g.generate_name('expand_result')
        nodes = [
            onnx.helper.make_node("Expand", [g.transients[0].name, g.transients[1].name], [expand_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(expand_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _expand



def flatten(axis=1):
    def _flatten(g):        
        flatten_result_name = g.generate_name('flatten_result')
        nodes = [
            onnx.helper.make_node("Flatten", [g.transients[0].name], [flatten_result_name], axis=axis),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(flatten_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _flatten


def gather_elements(axis=0):
    def _gather_elements(g):
        gather_elements_result_name = g.generate_name('gather_elements_result')        
        nodes = [
            onnx.helper.make_node(
                "GatherElements",
                [g.transients[0].name, g.transients[1].name],
                [gather_elements_result_name],
                axis=axis,
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(gather_elements_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _gather_elements


def gather_nd():
    """
    Input transients:
        - scores, as a 2D matrix
        - indices, as a [None, 2] matrix
    """
    def _gather_nd(g):
        gather_nd_result_name = g.generate_name('gather_nd_result')        
        nodes = [
            onnx.helper.make_node("GatherND", [g.transients[0].name, g.transients[1].name], [gather_nd_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(gather_nd_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _gather_nd


def identity(name, suffix=True):
    def _identity(g):        
        identity_name = g.generate_name(name) if suffix else name
        nodes = [
            onnx.helper.make_node("Identity", [g.transients[0].name], [identity_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(identity_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _identity


def less():
    def _less(g):        
        less_result_name = g.generate_name('less_result')        
        nodes = [
            onnx.helper.make_node("Less", [g.transients[0].name, g.transients[1].name], [less_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(less_result_name, onnx.TensorProto.BOOL, []),
            ],
        )

    return _less


def less_or_equal():
    def _less_or_equal(g):        
        less_or_equal_result_name = g.generate_name('less_or_equal_result')
        nodes = [
            onnx.helper.make_node("LessOrEqual", [g.transients[0].name, g.transients[1].name], [less_or_equal_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(less_or_equal_result_name, onnx.TensorProto.BOOL, []),
            ],
        )

    return _less_or_equal

def mul():
    def _mul(g):        
        mul_result_name = g.generate_name('mul_result')        
        nodes = [
            onnx.helper.make_node("Mul", [g.transients[0].name, g.transients[1].name], [mul_result_name]),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(mul_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )


    return _mul


def reduce_sum(keepdims=1, noop_with_empty_axes=0):
    def _reduce_sum(g):        
        reduce_sum_result_name = g.generate_name('reduce_sum_result')        
        nodes = [
            onnx.helper.make_node(
                "ReduceSum",
                [g.transients[0].name, g.transients[1].name],
                [reduce_sum_result_name],
                keepdims=keepdims,
                noop_with_empty_axes=noop_with_empty_axes,
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(reduce_sum_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )


    return _reduce_sum


def reshape(allowzero=0):
    def _reshape(g):        
        reshape_result_name = g.generate_name('reshape_result')        
        nodes = [
            onnx.helper.make_node(
                "Reshape",
                [g.transients[0].name, g.transients[1].name],
                [reshape_result_name],
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(reshape_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _reshape


def softmax(axis=-1):
    def _softmax(g):
        softmax_result_name = g.generate_name('softmax_result')        
        nodes = [
            onnx.helper.make_node(
                "Softmax",
                [g.transients[0].name],
                [softmax_result_name],
                axis=axis,
            ),
        ]

        return g._replace(
            nodes=graph.extend(g.nodes, nodes),
            transients=[
                 onnx.helper.make_tensor_value_info(softmax_result_name, onnx.TensorProto.UNDEFINED, []),
            ],
        )

    return _softmax



import ebm2onnx.graph as graph
import onnx


def test_create_graph():
    g = graph.create_graph()

    assert g.generate_name is not None
    assert g.inputs == []
    assert g.outputs == []
    assert g.nodes == []


def test_create_one_input():
    g = graph.create_graph()

    input = graph.create_input(g, "foo", onnx.TensorProto.FLOAT, [None, 3])
    assert len(input.inputs) == 1
    assert input.inputs == [onnx.helper.make_tensor_value_info(
        'foo' ,
        onnx.TensorProto.FLOAT,
        [None, 3])
    ]
    assert input.inputs == input.transients


def test_create_several_inputs():
    g = graph.create_graph()

    i1 = graph.create_input(g, "foo", onnx.TensorProto.FLOAT, [None, 3])
    i2 = graph.create_input(g, "bar", onnx.TensorProto.INT64, [None, 2])

    assert i1.inputs == [onnx.helper.make_tensor_value_info(
        'foo' ,
        onnx.TensorProto.FLOAT,
        [None, 3])
    ]
    assert i1.inputs == i1.transients

    assert i2.inputs == [onnx.helper.make_tensor_value_info(
        f'bar' ,
        onnx.TensorProto.INT64,
        [None, 2])
    ]
    assert i2.inputs == i2.transients


def test_create_initializer():
    g = graph.create_graph()

    init = graph.create_initializer(g, "foo", onnx.TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4])
    assert len(init.initializers) == 1
    assert init.initializers == [onnx.helper.make_tensor(
        'foo_0' ,
        onnx.TensorProto.FLOAT,
        [4],
        [0.1, 0.2, 0.3, 0.4]
    )]
    assert init.initializers == init.transients


def test_merge():
    g = graph.create_graph()

    init1 = graph.create_initializer(g, "foo", onnx.TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4])
    init2 = graph.create_initializer(g, "foo", onnx.TensorProto.FLOAT, [4], [1.1, 1.2, 3.3, 4.4])
    input1 = graph.create_input(g, "bar1", onnx.TensorProto.FLOAT, [None, 3])
    input2 = graph.create_input(g, "bar2", onnx.TensorProto.FLOAT, [None, 4])

    m = graph.merge(init1, input1, init2, input2)

    assert len(m.initializers) == 2
    assert len(m.inputs) == 2
    assert len(m.transients) == 4

    assert m.initializers == [
        onnx.helper.make_tensor(
            'foo_0' ,
            onnx.TensorProto.FLOAT,
            [4],
            [0.1, 0.2, 0.3, 0.4]
        ),
        onnx.helper.make_tensor(
            'foo_1' ,
            onnx.TensorProto.FLOAT,
            [4],
            [1.1, 1.2, 3.3, 4.4]
        ),
    ]

    assert m.inputs == [
        onnx.helper.make_tensor_value_info(
            'bar1' ,
            onnx.TensorProto.FLOAT,
            [None, 3],
        ),
        onnx.helper.make_tensor_value_info(
            'bar2' ,
            onnx.TensorProto.FLOAT,
            [None, 4],
        ),
    ]

    assert m.transients == [
        onnx.helper.make_tensor(
            'foo_0' ,
            onnx.TensorProto.FLOAT,
            [4],
            [0.1, 0.2, 0.3, 0.4]
        ),
        onnx.helper.make_tensor_value_info(
            'bar1' ,
            onnx.TensorProto.FLOAT,
            [None, 3],
        ),
        onnx.helper.make_tensor(
            'foo_1' ,
            onnx.TensorProto.FLOAT,
            [4],
            [1.1, 1.2, 3.3, 4.4]
        ),
        onnx.helper.make_tensor_value_info(
            'bar2' ,
            onnx.TensorProto.FLOAT,
            [None, 4],
        ),
    ]


def test_strip_to_transients():
    g = graph.create_graph()

    input1 = graph.create_input(g, "bar1", onnx.TensorProto.FLOAT, [None, 3])
    input2 = graph.create_input(g, "bar2", onnx.TensorProto.FLOAT, [None, 4])

    m = graph.merge(input1, input2)
    m = graph.strip_to_transients(m)

    assert m.initializers == []
    assert m.inputs == []
    assert m.transients == [
        onnx.helper.make_tensor_value_info(
            'bar1' ,
            onnx.TensorProto.FLOAT,
            [None, 3],
        ),
        onnx.helper.make_tensor_value_info(
            'bar2' ,
            onnx.TensorProto.FLOAT,
            [None, 4],
        ),
    ]


def test_clear_transients():
    g = graph.create_graph()

    input1 = graph.create_input(g, "bar1", onnx.TensorProto.FLOAT, [None, 3])
    input2 = graph.create_input(g, "bar2", onnx.TensorProto.FLOAT, [None, 4])

    m = graph.merge(input1, input2)
    m = graph.clear_transients(m)

    assert m.initializers == []
    assert m.inputs == [
        onnx.helper.make_tensor_value_info(
            'bar1' ,
            onnx.TensorProto.FLOAT,
            [None, 3],
        ),
        onnx.helper.make_tensor_value_info(
            'bar2' ,
            onnx.TensorProto.FLOAT,
            [None, 4],
        ),
    ]
    assert m.transients == []



import ebm2onnx.ebm as ebm
import ebm2onnx.graph as graph
import ebm2onnx.operators as ops
import onnx
import numpy as np

from .utils import assert_model_result, infer_model


def test_get_bin_index_on_continuous_value():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 1])

    g = ebm.get_bin_index_on_continuous_value([np.NINF, np.NINF, 0.2, 0.7, 1.2, 4.3])(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None, 1])
    
    assert_model_result(g, 
        input={
            'i': [
                [1.3],
                [0.6999],
                [-9.6],
                [9.6],
            ]
        },
        expected_result=[[
            [4],
            [2],
            [1],
            [5],
        ]]
    )


def test_get_bin_index_on_categorical_value():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.STRING, [None, 1])

    g = ebm.get_bin_index_on_categorical_value({
        'foo': 1,
        'bar': 2,
        'biz': 3,
    })(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None, 1])

    result = infer_model(graph.compile(g, target_opset=13),
        input={
            'i': [["biz"], ["foo"], ["bar"], ["nan"], ["okif"]],
        }
    )
    expected_result=[
            [3],
            [1],
            [2],
            [0],
            [-1],
        ]
    assert len(expected_result) == len(result[0])
    for i, r in enumerate(expected_result):
        assert result[0][i] == r


def test_get_bin_score_1d():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.INT64, [None, 1])

    g = ebm.get_bin_score_1d(np.array([0.0, 0.1, 0.2, 0.3]))(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 1, 1])
    
    assert_model_result(g, 
        input={
            'i': [
                [3],
                [1],
                [2],
                [0],
            ]
        },
        expected_result=[[
            [[0.3]],
            [[0.1]],
            [[0.2]],
            [[0.0]],
        ]]
    )


def test_get_bin_score_1d_multiclass():
    """test on 3 classes
    shape of scores is [bin_count x class_count]
    """
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.INT64, [None, 1])

    g = ebm.get_bin_score_1d(np.array(

        [
            [0.0, 1.0, 2.0],
            [0.1, 1.1, 2.1],
            [0.2, 1.2, 2.2],
            [0.3, 1.3, 2.3],
        ]
    ))(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 1, 3])

    assert_model_result(g,
        input={
            'i': [
                [3],
                [1],
                [2],
                [0],
                [2],
            ]
        },
        expected_result=[[
            [[0.3, 1.3, 2.3]],
            [[0.1, 1.1, 2.1]],
            [[0.2, 1.2, 2.2]],
            [[0.0, 1.0, 2.0]],
            [[0.2, 1.2, 2.2]],
        ]],
    )

def test_get_bin_score_2d():
    g = graph.create_graph()
    i1 = graph.create_input(g, "i1", onnx.TensorProto.INT64, [None, 1])
    i2 = graph.create_input(g, "i2", onnx.TensorProto.INT64, [None, 1])

    i = graph.merge(i1, i2)
    g = ebm.get_bin_score_2d(np.array([
        [0.0, 0.1, 0.2, 0.3],
        [1.0, 2.1, 3.2, 4.3],
        [10.0, 20.1, 30.2, 40.3],
    ]))(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 1, 1])
    
    assert_model_result(g, 
        input={
            'i1': [[2], [1], [2], [0]],
            'i2': [[3], [0], [2], [1]],
        },
        expected_result=[[
            [[40.3]],
            [[1.0]],
            [[30.2]],
            [[0.1]],
        ]]
    )


def test_compute_class_score():
    g = graph.create_graph()
    i1 = graph.create_input(g, "i1", onnx.TensorProto.FLOAT, [None, 1, 1])
    i2 = graph.create_input(g, "i2", onnx.TensorProto.FLOAT, [None, 1, 1])
    i3 = graph.create_input(g, "i3", onnx.TensorProto.FLOAT, [None, 1, 1])

    i = graph.merge(i1, i2, i3)
    g, _ = ebm.compute_class_score(np.array([0.2]), explain_name="scores")(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 1])
    
    assert_model_result(g, 
        input={
            'i1': [[[0.1]], [[0.2]], [[0.3]], [[0.4]]],
            'i2': [[[1.1]], [[1.2]], [[1.3]], [[1.4]]],
            'i3': [[[2.1]], [[2.2]], [[2.3]], [[2.4]]],
        },
        expected_result=[[
            [3.5],
            [3.8],
            [4.1],
            [4.4],
        ]]
    )


def test_compute_multiclass_score():
    g = graph.create_graph()
    i1 = graph.create_input(g, "i1", onnx.TensorProto.FLOAT, [None, 1, 3])
    i2 = graph.create_input(g, "i2", onnx.TensorProto.FLOAT, [None, 1, 3])
    i3 = graph.create_input(g, "i3", onnx.TensorProto.FLOAT, [None, 1, 3])

    i = graph.merge(i1, i2, i3)
    g, _ = ebm.compute_class_score(np.array([0.1, 0.2, 0.3]), explain_name="scores")(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 3])

    assert_model_result(g,
        input={
            'i1': [
                [[0.1, 0.2, 0.3]],
                [[0.2, 0.3, 0.4]],
                [[0.3, 0.4, 0.5]],
                [[0.4, 0.5, 0.6]]],
            'i2': [
                [[1.1, 1.2, 1.3]],
                [[1.2, 1.3, 1.4]],
                [[1.3, 1.4, 1.5]],
                [[1.4, 1.5, 1.6]]],
            'i3': [
                [[2.1, 2.2, 2.3]],
                [[2.2, 2.3, 2.4]],
                [[2.3, 2.4, 2.5]],
                [[2.4, 2.5, 2.6]]],
        },
        expected_result=[[
            [3.4, 3.8, 4.2],
            [3.7, 4.1, 4.5],
            [4.0, 4.4, 4.8],
            [4.3, 4.7, 5.1],
        ]]
    )


def test_predict_class_binary():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 1])

    g = ebm.predict_class(binary=True, prediction_name="prediction")(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])
    
    assert_model_result(g, 
        input={
            'i': [[3.5], [-3.8], [-0.1], [0.2]]
        },
        expected_result=[[1, 0, 0, 1]]
    )


def test_predict_multiclass_binary():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 3])

    g = ebm.predict_class(binary=False, prediction_name="prediction")(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])

    assert_model_result(g,
        input={
            'i': [
            [3.4, 3.8, 4.2],
            [3.7, 4.1, 0.5],
            [4.0, 0.4, 0.8],
            [4.3, 4.7, 5.1],
        ]
        },
        expected_result=[[2, 1, 0, 2]]
    )


def test_predict_proba_binary():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 1])

    g = ebm.predict_proba(binary=True, probabilities_name="probabilities")(i)
    g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, 2])
    
    assert_model_result(g, 
        input={
            'i': [[3.5], [-3.8], [-0.1], [0.2]]
        },
        expected_result=[[
            [0.02931223, 0.97068775],
            [0.97811866, 0.02188127],
            [0.5249792 , 0.4750208 ],
            [0.450166  , 0.54983395],
        ]]
    )




from onnx import defs


def get_latest_opset_version():
    """
    This module relies on *onnxruntime* to test every
    converter. The function returns the most recent
    target opset tested with *onnxruntime* or the opset
    version specified by *onnx* package if this one is lower
    (return by `onnx.defs.onnx_opset_version()`).
    """
    return min(13, defs.onnx_opset_version())



import os
import pandas as pd
import ebm2onnx


def test_titanic_types():
    df = pd.read_csv(os.path.join('examples','titanic_train.csv'))

    df['Survived'] = df['Survived'].astype(bool)
    df['Fare'] = df['Fare'].astype('float32')

    assert ebm2onnx.get_dtype_from_pandas(df) == {
        'PassengerId': 'int',
        'Survived': 'bool',
        'Pclass': 'int',
        'Name': 'str',
        'Sex': 'str',
        'Age': 'double',
        'SibSp': 'int',
        'Parch': 'int',
        'Ticket': 'str',
        'Fare': 'float',
        'Cabin': 'str',
        'Embarked': 'str',
    }



import ebm2onnx.graph as graph
import ebm2onnx.operators as ops

import numpy as np
import onnx

from .utils import assert_model_result


def test_add():
    g = graph.create_graph()

    a = graph.create_initializer(g, "a", onnx.TensorProto.FLOAT, [1], [0.3])
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None])

    l = ops.add()(graph.merge(i, a))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None])
    
    assert_model_result(l, 
        input={
            'i': [0.1, 1.2, 11, 4.2],
        },
        expected_result=[[0.4, 1.5, 11.3, 4.5]]
    )


def test_cast():
    g = graph.create_graph()

    i = graph.create_input(g, "i", onnx.TensorProto.INT64, [None, 1])

    l = ops.cast(onnx.TensorProto.FLOAT)(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 1])
    
    assert_model_result(l, 
        input={
            'i': [
                [1],
                [2],
                [11],
                [4],
            ]
        },
        expected_result=[[
            [1.0],
            [2.0],
            [11.0],
            [4.0]
        ]]
    )


def test_flatten():
    g = graph.create_graph()

    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None])

    l = ops.flatten()(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 1])
    
    assert_model_result(l, 
        input={
            'i': [0.1, 0.2, 0.3, 0.4]
        },
        expected_result=[[
            [0.1],
            [0.2],
            [0.3],
            [0.4]
        ]]
    )


def test_less():
    g = graph.create_graph()

    a = graph.create_initializer(g, "a", onnx.TensorProto.FLOAT, [4], [1.1, 2.3, 3.5, 9.6])
    b = graph.create_input(g, "b", onnx.TensorProto.FLOAT, [None, 1])

    l = ops.less()(graph.merge(a, b))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.BOOL, [None, 4])
    
    assert_model_result(l, 
        input={
            'b': [
                [0.1],
                [1.2],
                [11],
                [4.2],
                [np.NaN],
            ]
        },
        expected_result=[[
            [False, False, False, False],
            [True, False, False, False],
            [True, True, True, True],
            [True, True, True, False],
            [False, False, False, False],
        ]]
    )


def test_mul():
    g = graph.create_graph()

    a = graph.create_initializer(g, "a", onnx.TensorProto.FLOAT, [3], [1.0, 2.0, 3.0])
    b = graph.create_input(g, "b", onnx.TensorProto.FLOAT, [None, 3])

    l = ops.mul()(graph.merge(a, b))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 3])
    
    assert_model_result(l, 
        input={
            'b': [
                [0.1, 0.1, 0.1],
                [0.1, 0.2, 0.3],
            ]
        },
        expected_result=[[
            [0.1, 0.2, 0.3],
            [0.1, 0.4, 0.9],
        ]]
    )


def test_argmax():
    g = graph.create_graph()
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 3])

    l = ops.argmax(axis=1)(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.INT64, [None, 1])
    
    assert_model_result(l, 
        input={
            'i': [
                [1, 4, 2],
                [2, 8, 12],
                [11, 0, 5],
            ]
        },
        expected_result=[[
            [1],
            [2],
            [0],
        ]]
    )


def test_gather_elements():
    g = graph.create_graph()

    a = graph.create_initializer(g, "a", onnx.TensorProto.FLOAT, [3, 1], [0.1, 0.2, 0.3])
    b = graph.create_input(g, "b", onnx.TensorProto.INT64, [None, 1])

    l = ops.gather_elements()(graph.merge(a, b))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 1])
    
    assert_model_result(l, 
        input={
            'b': [
                [2],
                [1],
                [0],
            ]
        },
        expected_result=[[
            [0.3],
            [0.2],
            [0.1],
        ]]
    )


def test_gather_nd():
    g = graph.create_graph()

    a = graph.create_initializer(g, "a", onnx.TensorProto.FLOAT, [3, 3], np.array([
        [0.1, 0.2, 0.3],
        [1.1, 2.2, 3.3],
        [0.1, 20.2, 30.3],
    ]).flatten())
    b = graph.create_input(g, "b", onnx.TensorProto.INT64, [None, 2])

    l = ops.gather_nd()(graph.merge(a, b))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None])
    
    assert_model_result(l, 
        input={
            'b': [
                [2, 0],
                [1, 1],
                [0, 1],
            ]
        },
        expected_result=np.array([[0.1, 2.2, 0.2]])
    )


def test_concat():
    g = graph.create_graph()

    a = graph.create_input(g, "a", onnx.TensorProto.FLOAT, [3, 1])
    b = graph.create_input(g, "b", onnx.TensorProto.FLOAT, [3, 1])

    l = ops.concat(axis=1)(graph.merge(a, b))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 2])
    
    assert_model_result(l, 
        input={
            'a': [[0.1], [0.2], [0.3]],
            'b': [[1.1], [1.2], [1.3]],
        },
        expected_result=[[
            [0.1, 1.1],
            [0.2, 1.2],
            [0.3, 1.3],
        ]]
    )


def test_expand():
    g = graph.create_graph()

    shape = graph.create_initializer(g, "shape", onnx.TensorProto.INT64, [2], [4, 3])
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 1])

    l = ops.expand()(graph.merge(i, shape))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 3])

    assert_model_result(l,
        input={
            'i': [
                [0.1],
                [1.2],
                [11],
                [4.2],
            ]
        },
        expected_result=[[
            [0.1, 0.1, 0.1],
            [1.2, 1.2, 1.2],
            [11, 11, 11],
            [4.2, 4.2, 4.2]
        ]],
    )



def test_reduce_sum():
    g = graph.create_graph()

    axis = graph.create_initializer(g, "axis", onnx.TensorProto.INT64, [1], [1])
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 3])

    l = ops.reduce_sum(keepdims=0)(graph.merge(i, axis))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None])
    
    assert_model_result(l, 
        input={
            'i': [
                [0.1, 1.0, 1.2],
                [1.2, 0.4, 0.9],
                [11, 0.8, -0.2],
                [4.2, 3.2, -6.4],
            ]
        },
        expected_result=[[2.3, 2.5, 11.6, 1.0]]
    )


def test_reshape():
    g = graph.create_graph()

    shape = graph.create_initializer(g, "shape", onnx.TensorProto.INT64, [1], [0])
    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 1])

    l = ops.reshape()(graph.merge(i, shape))
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None])
    
    assert_model_result(l, 
        input={
            'i': [
                [0.1],
                [1.2],
                [11],
                [4.2],
            ]
        },
        expected_result=[[0.1, 1.2, 11, 4.2]]
    )


def test_softmax():
    g = graph.create_graph()

    i = graph.create_input(g, "i", onnx.TensorProto.FLOAT, [None, 2])

    l = ops.softmax()(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.FLOAT, [None, 2])
    
    assert_model_result(l, 
        input={
            'i': [
                [0.0, 0.68],
                [0.0, 0.2],
                [1.2, 0.3],
                [0.0, -0.2],
            ]
        },
        expected_result=[[
            [0.3362613 , 0.66373867],
            [0.450166  , 0.54983395],
            [0.7109495 , 0.2890505 ],
            [0.54983395, 0.450166  ]
        ]],
    )



import os
import pytest
import numpy as np
import pandas as pd

from interpret.glassbox import ExplainableBoostingClassifier, ExplainableBoostingRegressor
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder

import onnx
import ebm2onnx
from .utils import infer_model, create_session


def train_titanic_binary_classification(interactions=0, with_categorical=False):
    df = pd.read_csv(
        os.path.join('examples','titanic_train.csv'),
        #dtype= {
        #    'Age': np.float32,
        #    'Fare': np.float32,
        #    'Pclass': np.float32, # np.int
        #}
    )
    df = df.dropna()
    df['Old'] = df['Age'] > 65
    if with_categorical is False:
        feature_types=['continuous', 'continuous', 'continuous', 'continuous']
        feature_columns = ['Age', 'Fare', 'Pclass', 'Old']
    else:
        feature_types=['continuous', 'continuous', 'nominal', 'continuous', 'nominal']
        feature_columns = ['Age', 'Fare', 'Pclass', 'Old', 'Embarked']
    label_column = "Survived"

    y = df[[label_column]]
    le = LabelEncoder()
    y_enc = le.fit_transform(y)
    x = df[feature_columns]
    x_train, x_test, y_train, y_test = train_test_split(x, y_enc)
    model = ExplainableBoostingClassifier(interactions=interactions, feature_types=feature_types)
    model.fit(x_train, y_train)

    return model, x_test, y_test


def train_titanic_regression(interactions):
    df = pd.read_csv(os.path.join('examples','titanic_train.csv'))
    df = df.dropna()
    feature_columns = ['SibSp', 'Fare', 'Pclass']
    label_column = "Age"

    y = df[[label_column]]
    le = LabelEncoder()
    y_enc = le.fit_transform(y)
    x = df[feature_columns]
    x_train, x_test, y_train, y_test = train_test_split(x, y_enc)
    model = ExplainableBoostingRegressor(interactions=interactions)
    model.fit(x_train, y_train)

    return model, x_test, y_test


def train_bank_churners_multiclass_classification():
    df = pd.read_csv(
        os.path.join('examples','BankChurners.csv'),
    )
    df = df.dropna()
    feature_types=['continuous', 'continuous', 'nominal', 'continuous']
    feature_columns = ['Customer_Age', 'Dependent_count', 'Education_Level', 'Credit_Limit']
    label_column = "Income_Category"

    y = df[[label_column]]
    le = LabelEncoder()
    y_enc = le.fit_transform(y)
    x = df[feature_columns]
    x_train, x_test, y_train, y_test = train_test_split(x, y_enc)
    model = ExplainableBoostingClassifier(interactions=0, feature_types=feature_types)
    model.fit(x_train, y_train)

    return model, x_test, y_test


@pytest.mark.parametrize("explain", [False, True])
@pytest.mark.parametrize("interactions", [0, 2, [(0, 1, 2)], [(0, 1, 2, 3)]])
def test_predict_binary_classification(interactions, explain):
    model_ebm, x_test, y_test = train_titanic_binary_classification(interactions=interactions)
    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=explain,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Age': x_test['Age'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
        'Old': x_test['Old'].values,
    })

    if explain is True:
        assert len(pred_onnx) == 2
    assert np.allclose(pred_ebm, pred_onnx[0])


@pytest.mark.parametrize("explain", [False, True])
@pytest.mark.parametrize("interactions", [0, 2, [(0, 1, 2)], [(0, 1, 2, 3)]])
def test_predict_proba_binary_classification(interactions, explain):
    model_ebm, x_test, y_test = train_titanic_binary_classification(interactions=interactions)
    pred_ebm = model_ebm.predict_proba(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        predict_proba=True,
        explain=explain,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
        }
    )
    pred_onnx = infer_model(model_onnx, {
        'Age': x_test['Age'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
        'Old': x_test['Old'].values,
    })

    if explain is True:
        assert len(pred_onnx) == 3        
    assert np.allclose(pred_ebm, pred_onnx[1])


@pytest.mark.parametrize("explain", [False, True])
@pytest.mark.parametrize("interactions", [0, 2, [(0, 1, 2)], [(0, 1, 2, 3)]])
def test_predict_regression_without_interactions(interactions, explain):
    model_ebm, x_test, y_test = train_titanic_regression(interactions=0)
    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=explain,
        dtype={
            'SibSp': 'int',
            'Fare': 'double',
            'Pclass': 'int',
        },
    )
    pred_onnx = infer_model(model_onnx, {
        'SibSp': x_test['SibSp'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
    })

    if explain is True:
        assert len(pred_onnx) == 2
    assert np.allclose(pred_ebm, pred_onnx[0])


@pytest.mark.parametrize("explain", [False, True])
@pytest.mark.parametrize("interactions", [0, 2, [(0, 1, 2)], [(0, 1, 2, 3)]])
def test_predict_binary_classification_with_categorical(interactions, explain):
    model_ebm, x_test, y_test = train_titanic_binary_classification(
        interactions=interactions,
        with_categorical=True,
    )
    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=explain,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
            'Embarked': 'str'
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Age': x_test['Age'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
        'Old': x_test['Old'].values,
        'Embarked': x_test['Embarked'].values,
    })

    if explain is True:
        assert len(pred_onnx) == 2
    assert np.allclose(pred_ebm, pred_onnx[0])


def test_predict_multiclass_classification():
    model_ebm, x_test, y_test = train_bank_churners_multiclass_classification()
    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        dtype={
            'Customer_Age': 'int',
            'Dependent_count': 'int',
            'Education_Level': 'str',
            'Credit_Limit': 'double',
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Customer_Age': x_test['Customer_Age'].values,
        'Dependent_count': x_test['Dependent_count'].values,
        'Education_Level': x_test['Education_Level'].values,
        'Credit_Limit': x_test['Credit_Limit'].values,
    })

    assert np.allclose(pred_ebm, pred_onnx[0])


def test_predict_proba_multiclass_classification():
    model_ebm, x_test, y_test = train_bank_churners_multiclass_classification()
    pred_ebm = model_ebm.predict_proba(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        predict_proba=True,
        explain=True,
        dtype={
            'Customer_Age': 'int',
            'Dependent_count': 'int',
            'Education_Level': 'str',
            'Credit_Limit': 'double',
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Customer_Age': x_test['Customer_Age'].values,
        'Dependent_count': x_test['Dependent_count'].values,
        'Education_Level': x_test['Education_Level'].values,
        'Credit_Limit': x_test['Credit_Limit'].values,
    })

    assert len(pred_onnx) == 3
    assert np.allclose(pred_ebm, pred_onnx[1])


def test_predict_w_scores_outputs_def():
    model_ebm, _, _ = train_titanic_binary_classification(interactions=0)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=True,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
        }
    )
    session = create_session(model_onnx)

    outputs = session.get_outputs()
    assert len(outputs) == 2
    assert outputs[0].name == "prediction"
    assert outputs[0].shape == [None]
    assert outputs[0].type == 'tensor(int64)'
    assert outputs[1].name == "scores"
    assert outputs[1].shape == [None, 4, 1]
    assert outputs[1].type == 'tensor(float)'


def test_predict_proba_w_scores_outputs_def():
    model_ebm, _, _ = train_titanic_binary_classification(interactions=0)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        predict_proba=True,
        explain=True,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
        }
    )
    session = create_session(model_onnx)

    outputs = session.get_outputs()
    assert len(outputs) == 3
    assert outputs[0].name == "prediction"
    assert outputs[0].shape == [None]
    assert outputs[0].type == 'tensor(int64)'
    assert outputs[1].name == "probabilities"
    assert outputs[1].shape == [None, 2]
    assert outputs[1].type == 'tensor(float)'
    assert outputs[2].name == "scores"
    assert outputs[2].shape == [None, 4, 1]
    assert outputs[2].type == 'tensor(float)'



def test_predict_binary_classification_missing_values():
    model_ebm, x_test, y_test = train_titanic_binary_classification(with_categorical=True)

    # patch data
    x_test.iloc[0, x_test.columns.get_loc('Age')] = np.nan
    x_test.iloc[1, x_test.columns.get_loc('Fare')] = np.nan
    x_test.iloc[2, x_test.columns.get_loc('Embarked')] = np.nan

    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=True,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
            'Embarked': 'str'
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Age': x_test['Age'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
        'Old': x_test['Old'].values,
        'Embarked': x_test['Embarked'].values,
    })
    
    assert np.allclose(pred_ebm, pred_onnx[0])

    # score of NaN Age on line 0 must be 0
    assert pred_onnx[1][0][0][0] == 0.0  # index: score,iloc,Age, 0

    # score of NaN Fare on line 1 must be 0
    assert pred_onnx[1][1][1][0] == 0.0  # index: score,iloc,Fare, 0

    # score of NaN Embarked on line 2 must be 0
    assert pred_onnx[1][2][4][0] == 0.0  # index: score,iloc,Embarked, 0


def test_predict_binary_classification_unknown_values():
    model_ebm, x_test, y_test = train_titanic_binary_classification(with_categorical=True)

    # patch data
    x_test.iloc[0, x_test.columns.get_loc('Pclass')] = 5
    x_test.iloc[1, x_test.columns.get_loc('Pclass')] = -2
    x_test.iloc[2, x_test.columns.get_loc('Embarked')] = 'Z'

    pred_ebm = model_ebm.predict(x_test)

    model_onnx = ebm2onnx.to_onnx(
        model_ebm,
        explain=True,
        dtype={
            'Age': 'double',
            'Fare': 'double',
            'Pclass': 'int',
            'Old': 'bool',
            'Embarked': 'str'
        }
    )

    pred_onnx = infer_model(model_onnx, {
        'Age': x_test['Age'].values,
        'Fare': x_test['Fare'].values,
        'Pclass': x_test['Pclass'].values,
        'Old': x_test['Old'].values,
        'Embarked': x_test['Embarked'].values,
    })
    
    assert np.allclose(pred_ebm, pred_onnx[0])

    # score of Pclass on line 0 must be 0
    assert pred_onnx[1][0][2][0] == 0.0  # index: score,iloc,Pclass, 0

    # score of Pclass on line 1 must be 0
    assert pred_onnx[1][1][2][0] == 0.0  # index: score,iloc,Pclass, 0

    # score of Embarked on line 2 must be 0
    assert pred_onnx[1][2][4][0] == 0.0  # index: score,iloc,Embarked, 0



from collections import namedtuple
from .utils import get_latest_opset_version
from ebm2onnx import graph
from ebm2onnx import ebm
import ebm2onnx.operators as ops

import numpy as np
import onnx

from interpret.glassbox import ExplainableBoostingClassifier, ExplainableBoostingRegressor


onnx_type_for={
    'bool': onnx.TensorProto.BOOL,
    'float': onnx.TensorProto.FLOAT,
    'double': onnx.TensorProto.DOUBLE,
    'int': onnx.TensorProto.INT64,
    'str': onnx.TensorProto.STRING,
}


def infer_features_dtype(dtype, feature_name):
    feature_dtype = onnx.TensorProto.DOUBLE
    if dtype is not None:
        feature_dtype = onnx_type_for[dtype[feature_name]]

    return feature_dtype


def get_dtype_from_pandas(df):
    """Infers the features names and types from a pandas dataframe

    Example:
        >>>import ebm2onnx
        >>>
        >>>dtype = ebm2onnx.get_dtype_from_pandas(my_df)
    Args:
        df: A pandas dataframe

    Returns:
        A dict that can be used as the type argument of the to_onnx function.
    """
    dtype = {}
    df_types = df.dtypes.values
    for i, k in enumerate(df.dtypes.index):
        if df_types[i] == np.float32:
            dtype[k] = 'float'
        elif df_types[i] == np.double:
            dtype[k] = 'double'
        elif df_types[i] == int:
            dtype[k] = 'int'
        elif df_types[i] == bool:
            dtype[k] = 'bool'
        elif df_types[i] == str:
            dtype[k] = 'str'
        elif df_types[i] == object:
            dtype[k] = 'str'
        else:
            raise ValueError("column {} is of type {} that is not supported".format(k, df_types[i]))

    return dtype


def to_onnx(model, dtype, name="ebm",
            predict_proba=False,
            explain=False,
            target_opset=None,
            prediction_name="prediction",
            probabilities_name="probabilities",
            explain_name="scores",
            ):
    """Converts an EBM model to ONNX.

    The returned model contains one to three output.
    The first output is always the prediction, and is named "prediction".
    If predict_proba is set to True, then another output named "probabilities" is added.
    If explain is set to True, then another output named "scores" is added.

    Args:
        model: The EBM model, trained with interpretml
        dtype: A dict containing the type of each input feature. Types are expressed as strings, the following values are supported: float, double, int, str.
        name: [Optional] The name of the model
        predict_proba: [Optional] For classification models, output prediction probabilities instead of class
        explain: [Optional] Adds an additional output with the score per feature per class
        target_opset: [Optional] The target onnx opset version to use

    Returns:
        An ONNX model.
    """
    target_opset = target_opset or get_latest_opset_version()
    root = graph.create_graph()

    class_index=0
    inputs = [None for _ in model.feature_names_in_]
    parts = []

    feature_types = list(model.feature_types_in_)
    interaction_count = len(model.term_names_) - len(feature_types)
    for _ in range(interaction_count):
        feature_types.append('interaction')

    # first compute the score of each feature
    for feature_index in range(len(model.term_names_)):
        feature_name=model.term_names_[feature_index]
        feature_type=feature_types[feature_index]
        feature_group=model.term_features_[feature_index]

        if feature_type == 'continuous':
            bins = [np.NINF, np.NINF] + list(model.bins_[feature_group[0]][0])
            additive_terms = model.term_scores_[feature_index]

            feature_dtype = infer_features_dtype(dtype, feature_name)
            part = graph.create_input(root, feature_name, feature_dtype, [None])
            part = ops.flatten()(part)
            inputs[feature_index] = part
            part = ebm.get_bin_index_on_continuous_value(bins)(part)
            part = ebm.get_bin_score_1d(additive_terms)(part)
            parts.append(part)

        elif feature_type in ['nominal', 'ordinal']:
            col_mapping = model.bins_[feature_group[0]][0]
            additive_terms = model.term_scores_[feature_index]

            feature_dtype = infer_features_dtype(dtype, feature_name)
            part = graph.create_input(root, feature_name, feature_dtype, [None])
            if feature_dtype != onnx.TensorProto.STRING:
                part = ops.cast(onnx.TensorProto.STRING)(part)
            part = ops.flatten()(part)
            inputs[feature_index] = part
            part = ebm.get_bin_index_on_categorical_value(col_mapping)(part)
            part = ebm.get_bin_score_1d(additive_terms)(part)
            parts.append(part)

        elif feature_type == 'interaction':
            i_parts = []
            way_count = len(feature_group)

            for index in range(way_count):
                i_feature_index = feature_group[index]
                i_feature_type = feature_types[i_feature_index]

                if i_feature_type == 'continuous':
                    # interactions can be of any size (n way).
                    # There may be one binning per interaction way or not.
                    # the rule is to use bins_ index if there is one binning available for the way count.
                    # otherwise, use the last binning for the feature
                    bin_index = -1 if way_count > len(model.bins_[i_feature_index]) else way_count - 1
                    bins = [np.NINF, np.NINF] + list(model.bins_[i_feature_index][bin_index])
                    input = graph.strip_to_transients(inputs[i_feature_index])
                    i_parts.append(ebm.get_bin_index_on_continuous_value(bins)(input))

                elif i_feature_type in ['nominal', 'ordinal']:
                    col_mapping = model.bins_[i_feature_index][0]
                    input = graph.strip_to_transients(inputs[i_feature_index])
                    i_parts.append(ebm.get_bin_index_on_categorical_value(col_mapping)(input))

                else:
                    raise ValueError(f"The type of the feature {feature_name} is unknown: {feature_type}")

            part = graph.merge(*i_parts)
            additive_terms = model.term_scores_[feature_index]
            part = ebm.get_bin_score_2d(np.array(additive_terms))(part)
            parts.append(part)

        else:
            raise ValueError(f"The type of the feature {feature_name} is unknown: {feature_type}")

    # compute scores, predict and proba
    g = graph.merge(*parts)
    if type(model) is ExplainableBoostingClassifier:
        g, scores_output_name = ebm.compute_class_score(model.intercept_, explain_name)(g)
        g_scores = graph.strip_to_transients(g)
        if len(model.classes_) == 2: # binary classification            
            g = ebm.predict_class(binary=True, prediction_name=prediction_name)(g)
            g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])
            if predict_proba is True:
                gp = ebm.predict_proba(binary=True, probabilities_name=probabilities_name)(g_scores)
                g = graph.merge(graph.clear_transients(g), gp)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, len(model.classes_)])
        else:
            g = ebm.predict_class(binary=False, prediction_name=prediction_name)(g)
            g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.INT64, [None])
            if predict_proba is True:
                gp = ebm.predict_proba(binary=False, probabilities_name=probabilities_name)(g_scores)
                g = graph.merge(graph.clear_transients(g), gp)
                g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None, len(model.classes_)])

        if explain is True:
            if len(model.classes_) == 2:
                g = graph.add_output(g, scores_output_name, onnx.TensorProto.FLOAT, [None, len(model.term_names_), 1])
            else:
                g = graph.add_output(g, scores_output_name, onnx.TensorProto.FLOAT, [None, len(model.term_names_), len(model.classes_)])
    elif type(model) is ExplainableBoostingRegressor:
        g, scores_output_name = ebm.compute_class_score(np.array([model.intercept_]), explain_name)(g)
        g = ebm.predict_value(prediction_name)(g)
        g = graph.add_output(g, g.transients[0].name, onnx.TensorProto.FLOAT, [None])
        g = graph.add_output(g, scores_output_name, onnx.TensorProto.FLOAT, [None, len(model.term_names_), 1])
    else:
        raise NotImplementedError("{} models are not supported".format(type(model)))



    model = graph.compile(g, target_opset, name=name)
    return model



import ebm2onnx.graph as graph
import ebm2onnx.operators_ml as mlops

import numpy as np
import onnx

from .utils import assert_model_result, infer_model


def test_category_mapper_str2int():
    g = graph.create_graph()

    i = graph.create_input(g, "i", onnx.TensorProto.STRING, [None])

    l = mlops.category_mapper(
        cats_int64s=[0, 1, 2],
        cats_strings=["foo", "bar", "biz"],
    )(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.INT64, [None])
    
    result = infer_model(graph.compile(l, target_opset=13),
        input={
            'i': ["biz", "foo", "bar", "flah"],
        }
    )
    expected_result = [2, 0, 1, -1]
    assert len(expected_result) == len(result[0])
    for i, r in enumerate(expected_result):
        assert result[0][i] == r


def test_category_mapper_int2str():
    g = graph.create_graph()

    i = graph.create_input(g, "i", onnx.TensorProto.INT64, [None])

    l = mlops.category_mapper(
        cats_int64s=[0, 1, 2],
        cats_strings=["foo", "bar", "biz"],
    )(i)
    l = graph.add_output(l, l.transients[0].name, onnx.TensorProto.STRING, [None])
    
    result = infer_model(graph.compile(l, target_opset=13),
        input={
            'i': [2, 0, 1, 8],
        }
    )
    expected_result = ["biz", "foo", "bar", "_Unused"]
    assert len(expected_result) == len(result[0])
    for i, r in enumerate(expected_result):
        assert result[0][i] == r
